/*******************************************

 MAIN.CPP - The KINT Interpreter
	Copyright 2002-2003 Everett Kaser
	All rights reserved.

	Do not copy or redistribute without the
	author's explicit written permission.
 *******************************************

The KINT 2.0 interpreter is divided into two .C files and two matching .H files.
The two .C files are pretty much divided into mostly OS-dependant code (MAIN.C)
and OS-indendent code (PMACH.C).  Of course, there's going to be some cross-over
between, particularly depending upon the "native bitmap data format" on the
target machine.  With 2.0, KINT does ALL of it's own drawing of graphics, including
bitmaps, points, lines, rectangles, ovals, and text.  This is all done to memory
arrays that represent the bitmap data, and there is one of these that represents
the game's display output at any given time (kBuf).  The bitmap data is arranged
(and thus manipulated by the routines in PMACH.C) such that it's consistent with
Microsoft Window's DIB bitmaps, and thus can be easily pointed to and BitBlt'd
to the actual screen during the WM_PAINT message (any time the OS asks for the
game window to be redrawn).  If you want to port KINT to another OS, its native
bitmap format may be different, in which case it would be more efficient (in
terms of code execution) if those routines in PMACH.C were modified so that the
KINT memory bitmap data format matched your target OS's, and the routines that
drew into it would, of course, then have to be modified to read/write that different
format.

The KINT 2.0 interpreter has added some significant functionality to the graphics,
including clipping and rotation of bitmaps.  The bitmaps don't actually get
rotated, but rather all of the drawing routines access the bitmap memory blocks
(BMBs) via pointers and global variables that get set up at the beginning of
each routine based upon the current rotation state of the source and target
bitmaps (see SetupBMD() and SetupBMS() in PMACH.C).  These routines set a pointer
to the pixel that in the BMB that is the LOGICAL top-left corner (if the bitmap
werer NOT rotated).  So, for example, if the bitmap is rotated 180 degrees, then
this pointer would be set to the PHYSICAL lower-right pixel of the bitmap.  Further,
there are variables that give the amount (offset) to add to the pointer in order
to step one pixel in the X direction and one pixel in the Y direction.  Similarly,
a translation process is done in MAIN.C to convert mouse-pointer coordinates from
the physical screen coordinates to properly "rotated" coordinates (see FixXY() in
MAIN.C) and to convert keyboard cursor keys (see MapCursor() in MAIN.C) so that
everything works relatively transparently for the KASM code itself (ie, the KASM
.KE program doesn't have to know or care which way the bitmaps are rotated, although
it CAN find out using the SysSizeBM opcode and can rotate bitmaps using the SysOrientBM
opcode).  The BMB data structure has physW/physH and logW/logH values.  The phys values
are the actual width and height of the bitmap as it was allocated in RAM, whereas the
log values are the rotated-values which are dependant upon the bitmap's current
orientation at any given moment.

There is also the ability to have the graphics "double-sized".  This was
done it make developing Pocket PC applications on a desktop a little easier, since
a 320x240 window on a 1280x1024 display is pretty dinky.  Normally on a desktop
interpreter, you'd want KINT_DOUBLE_SIZE to be 0.  Of course, that double-sizing
effects the display OUTPUT as well as the mouse INPUT.  There is also a KINT_POCKET_PC_EMUL
#define that controls whether the interpreter is trying to emulate a Pocket PC
environment on a desktop screen or not.  ie, these are the supported combinations:
                   KINT_DESKTOP_WINDOWS KINT_POCKET_PC  KINT_POCKET_PC_EMUL  KINT_DOUBLE_SIZE
desktop						1				  0					  0					0
Pocket PC					0				  1					  0					0
desktop Pocket PC			1				  0					  1					0
desktop Pocket PC			1				  0					  1					1

See MAIN.H for all of these #define's.

There is a lot of #if'd code sprinkled throughout MAIN.C to be able to generate a
KINT 2.0 interpreter for either the desktop MS Windows OS's OR for the Pocket PC
devices running Microsoft Windows CE OS.  If you're porting this to a completely
different OS, the simplest thing might be to go through the code and hack out the
#if'd code that doesn't match your OS very well (ie, if you're porting to another
handheld, you'd get rid of any KINT_DESKTOP_WINDOWS code, and if you're porting to
another desktop OS, you'd get rid of any KINT_POCKET_PC code).

There is some UNICODE stuff.  The KASM code only deals with 8-bit characters.
However, the Pocket PC uses ONLY Unicode for file names, and other OS's may
move to that in the future.  Hence, the Widen() and Unwiden() functions to 
convert back and forth between 8 and 16 bit characters.

The KHDR_IDLE message that is being passed in the main message-loop: it's not
currently (and probably will remain) not terribly important.  I can't guarantee
that, but I try to put all "periodic activity" on the KHDR_TIMER call.  So, if
your OS structure makes the KHDR_IDLE call difficult to implement, you can PROBABLY
get by without it (but I make no guarantees).

The code is not as clean and neat as I'd like, but time, reality, and the desire for
fun rather than drudgery all exert a great deal of influence on the final appearance.

 ************************************************************/

//
// Mac TODO:
//
// * Finish implementing printing
//
// Mac nice-to-have:
//
// - String overrides - e.g. turn Ctrl-S into Cmd-S
//
// - Hero's Friends writes bmp data to the game folder.
//   Fix so that it writes elsewhere and the game can still find it.
//
// - Find way to disable/hide the shareware "make a copy" buttons/UI.
//

#define DEBUG_FILE 1
#define DEBUG_EVENTS 1
#define DEBUG_SOUND 1
#define DEBUG_PRINT 1

#define FAKE_SCREEN_SIZE	0
#define FAKE_SCREEN_SIZE_W	800
#define FAKE_SCREEN_SIZE_H	600

/*********************  Header Files  *********************/

#include <CoreFoundation/CoreFoundation.h>
#include <QuickTime/QuickTime.h>

#ifdef __cplusplus
extern "C" {
#endif
#include <mach/mach_time.h>
#include "main.h"
#ifdef __cplusplus
}
#endif

#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>
#include <deque>
#include <unistd.h>
#include <sys/stat.h>

using std::string;

/*********************  Local Structures  *********************/

typedef struct
{
	Movie		mMovie;
	Handle		mDataRef;
	bool		mLooping;
	bool		mPlaying;
	bool		mAutoDelete;
	bool		mIsMusic;
	QTCallBack	qtCallback;
} movieSound_t;
static std::deque<movieSound_t *> sSoundOrphans;

class KINTmsg
{
public:
	WORD mCmd;
	DWORD mArg1, mArg2, mArg3;

	KINTmsg (WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3)
	{
		mCmd = cmd; mArg1 = arg1; mArg2 = arg2; mArg3 = arg3;
	}
	
	~KINTmsg () {}
};
std::deque<KINTmsg *> KINTMsgQueue;

typedef struct
{
	int width;
	int height;
} windowSizes_t;

/*********************  Local Prototypes  *********************/

int wildicmp(const char *wild, const char *string);
bool CreateMacWindow (void);
string NewBuildPath(const char *pMem, BYTE inFolderType);
void PlaySoundFromFile (const char *inFilePath);
void StopSound_QT (movieSound_t *inSound);

//********************************************************************
extern PBMB		kBuf, pbFirst;

//************************************************************
// pmach GLOBALS
//TCHAR	PgmPath[512], TmpPath[512];
int		PhysicalScreenW, PhysicalScreenH, PhysicalScreenSizeIndex;
WORD	MusicPlaying;
//int		CurCur;
KE		mainKE;

/*******************  Global Variables ********************/

CFStringRef gAppName;

WindowRef gMacWindow;
bool gInBackground;
bool sIgnoreNextMouseUp;
volatile bool gInPrintLoop;

EventLoopTimerRef sTimerRef = NULL;
EventLoopTimerRef sIdleTimerRef = NULL;
DWORD sTimerPeriod;

BOOL InSetWindowPos;
BOOL ProgramReady;
BOOL WinHide = false;
BOOL AppHide = false;
BOOL DockHide = false;

#pragma mark -

// KINT requires VK key codes. This table will take a Mac keyCode and remap it to the closest VK equivalent.
static BYTE sMacKeyCodeToVK[128] =
{
	0x41, 0x53, 0x44, 0x46, 0x48, 0x47, 0x5a, 0x58, 0x43, 0x56, 0x00, 0x42, 0x51, 0x57, 0x45, 0x52,
	0x59, 0x54, 0x31, 0x32, 0x33, 0x34, 0x36, 0x35, 0xbb, 0x39, 0x37, 0xbd, 0x38, 0x30, 0xdd, 0x4f,
	0x55, 0xdb, 0x49, 0x50, 0x0d, 0x4c, 0x4a, 0xde, 0x4b, 0xba, 0xdc, 0xbc, 0xbf, 0x4e, 0x4d, 0xbe,
	0x09, 0x20, 0xc0, 0x08, 0x00, 0x1b, 0x00, 0x00, 0x10, 0x14, 0x12, 0x11, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x6e, 0x00, 0x6a, 0x00, 0x6b, 0x00, 0x90, 0x00, 0x00, 0x00, 0x6f, 0x0d, 0x00, 0x6d, 0x00,
	0x00, 0xbb, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x00, 0x68, 0x69, 0x00, 0x00, 0x00,
	0x74, 0x75, 0x76, 0x72, 0x77, 0x78, 0x00, 0x7a, 0x00, 0x2c, 0x00, 0x91, 0x00, 0x79, 0x00, 0x7b,
	0x00, 0x13, 0x2d, 0x24, 0x21, 0x2e, 0x73, 0x23, 0x71, 0x22, 0x70, 0x25, 0x27, 0x28, 0x26, 0x00
};

// These are used by the modifier-changed events
enum
{
	VK_SHIFT	= 0x10,
	VK_CONTROL	= 0x11,
	VK_MENU		= 0x12,
	VK_CAPITAL	= 0x14
};

//********************************************************************
void AddMsg(WORD cmd, DWORD arg1, DWORD arg2, DWORD arg3)
{
	if (!gMacWindow || !sIdleTimerRef)
	{
#if _DEBUG && DEBUG_EVENTS
		printf("AddMsg cmd: %d %d %d %d\n", cmd, arg1, arg2, arg3);
#endif
		KEmsg(&mainKE, cmd, arg1, arg2, arg3);
	}
	else
	{
		KINTmsg *newMsg = new KINTmsg(cmd, arg1, arg2, arg3);
		KINTMsgQueue.push_back(newMsg);

		// Wake up the idle timer so it can process the KINT events immediately
		SetEventLoopTimerNextFireTime(sIdleTimerRef, kEventDurationNoWait);
	}
}

//********************************************************************
void SendNextMsg()
{
	if (!KINTMsgQueue.empty())
	{
		KINTmsg *kintMsg = KINTMsgQueue.front();
#if _DEBUG && DEBUG_EVENTS
		printf("SendNextMsg cmd: %d %d %d %d\n", kintMsg->mCmd, kintMsg->mArg1, kintMsg->mArg2, kintMsg->mArg3);
#endif
		KEmsg (&mainKE, kintMsg->mCmd, kintMsg->mArg1, kintMsg->mArg2, kintMsg->mArg3);
		KINTMsgQueue.pop_front();
		delete kintMsg;
	}
}

void MacNotImplemented(void)
{
#if _DEBUG
	Debugger();
#endif
}

bool MacDoQuit(void)
{
	if( KEmsg(&mainKE, KHDR_QUERY_QUIT, 0, 0, 0) )
	{
		KEmsg(&mainKE, KHDR_SHUT_DOWN, 0, 0, 0);
		return true;
	}
	return false;
}

pascal OSStatus appCommandHandler (EventHandlerCallRef handlerRef, EventRef inEvent, void *userData)
{
    OSStatus status = eventNotHandledErr;
	UInt32 eventClass = GetEventClass (inEvent);
	UInt32 eventKind = GetEventKind (inEvent);
	
	if (eventClass == kEventClassCommand && eventKind == kEventProcessCommand)
	{
		HICommand command; 
		GetEventParameter (inEvent, kEventParamDirectObject, typeHICommand, NULL, sizeof(HICommand), NULL,  &command);

		switch (command.commandID)
		{
			case kHICommandPreferences:
				break;

			case kHICommandQuit:
				MacDoQuit();
				status = noErr;
				break;

			default:
				status = eventNotHandledErr;
		}
	}
	else if (eventClass == kEventClassApplication)
	{
		switch (eventKind)
		{
			case kEventAppActivated:
				gInBackground = false;
				break;
			case kEventAppDeactivated:
				gInBackground = true;
				break;
			case kEventAppHidden:
				AddMsg(KHDR_WINDOW_HIDDEN, 1, 0, 0);
				AppHide = true;
				break;
			case kEventAppShown:
				if (AppHide)
					AddMsg(KHDR_WINDOW_HIDDEN, 0, 0, 0);
				AppHide = false;
				break;
		}
	}
    return status;
}

pascal OSStatus windowEventHandler (EventHandlerCallRef inCaller, EventRef inEvent, void* inUserData)
{
	OSStatus result = eventNotHandledErr;

	UInt32 eventClass = GetEventClass (inEvent);
	UInt32 eventKind = GetEventKind (inEvent);
	
	SetPortWindowPort (gMacWindow);

	WindowRef macWindow;
	(void)GetEventParameter(inEvent, kEventParamWindowRef, typeWindowRef, NULL, sizeof(macWindow), NULL, &macWindow);

	if (eventClass == kEventClassWindow)
	{
		switch (eventKind)
		{
			case kEventWindowUpdate:
				CGrafPtr windowPort = GetWindowPort (gMacWindow);
				PixMapHandle spixmap = GetGWorldPixMap(kBuf->gWorld);

				Rect bounds;
				GetWindowPortBounds(gMacWindow, &bounds);

				ForeColor(blackColor);
				BackColor(whiteColor);

				LockPixels (spixmap);
				CopyBits (GetPortBitMapForCopyBits ((CGrafPtr)kBuf->gWorld), GetPortBitMapForCopyBits (windowPort),
					&bounds, &bounds, srcCopy, NULL);
				UnlockPixels (spixmap);
				break;
				
			// The window is going to/from the dock (no size change)
			case kEventWindowCollapsed:
				AddMsg(KHDR_WINDOW_HIDDEN, 1, 0, 0);
				DockHide = true;
				break;
			case kEventWindowExpanded:
				if (DockHide)
					AddMsg(KHDR_WINDOW_HIDDEN, 0, 0, 0);
				DockHide = false;
				break;

			case kEventWindowHidden:
				AddMsg(KHDR_WINDOW_HIDDEN, 1, 0, 0);
				WinHide = true;
				break;

			case kEventWindowShown:
				// KINT doesn't like receiving unhide messages at startup when the window is first shown.
				if (WinHide == true)
					AddMsg(KHDR_WINDOW_HIDDEN, 0, 0, 0);

				if (!InSetWindowPos)
				{
					if (kBuf->logW > kBuf->physW || kBuf->logH > kBuf->physH)
						InitScreen(kBuf->logW, kBuf->logH);
					AddMsg(KHDR_WINDOW_SIZE, kBuf->logW, kBuf->logH, K_MAXIMIZED);
				}
				WinHide = false;
				break;
				
			case kEventWindowBoundsChanged:
				{
					UInt32 attribs;
					(void)GetEventParameter(inEvent, kEventParamAttributes, typeUInt32, NULL, sizeof(attribs), NULL, &attribs);
					if (attribs & kWindowBoundsChangeSizeChanged)
					{
						Rect newBounds;
						(void)GetEventParameter(inEvent, kEventParamCurrentBounds, typeQDRectangle, NULL, sizeof(newBounds), NULL, &newBounds);
						
						kBuf->logW = newBounds.right - newBounds.left;
						kBuf->logH = newBounds.bottom - newBounds.top;

						if (!InSetWindowPos)
						{
							if (kBuf->logW > kBuf->physW || kBuf->logH > kBuf->physH)
								InitScreen(kBuf->logW, kBuf->logH);

							AddMsg(KHDR_WINDOW_SIZE, kBuf->logW, kBuf->logH, K_SIZED);
						}
					}
					result = noErr;
				}
				break;

			case kEventWindowClose:
				MacDoQuit();
				result = noErr;
				break;
		}
	}
	else if (eventClass == kEventClassKeyboard)
	{
		char keyChar;
		(void)GetEventParameter(inEvent, kEventParamKeyMacCharCodes, typeChar, NULL, sizeof(keyChar), NULL, &keyChar);
		UInt32 modifiers;
		(void)GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(modifiers), NULL, &modifiers);

		UInt32 keyCode;
		(void)GetEventParameter(inEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(keyCode), NULL, &keyCode);
		BYTE vkCode = sMacKeyCodeToVK[keyCode & 0x7f];
		
		switch (eventKind)
		{
			case kEventRawKeyDown:
				if( (vkCode >= 0x30 && vkCode <= 0x39) || (vkCode >= 0x41 && vkCode <= 0x5A) )
					AddMsg(KHDR_ALPHAKEY_DOWN, vkCode, 0 /* repeatRate */, 0);
				else
					AddMsg(KHDR_KEY_DOWN, vkCode, 0 /* repeatRate */, 0);
				AddMsg(KHDR_CHAR, keyChar, 0 /* repeatRate */, 0);
				result = noErr;
				break;
				
			case kEventRawKeyUp:
				if( (vkCode >= 0x30 && vkCode <= 0x39) || (vkCode >= 0x41 && vkCode <= 0x5A) )
					AddMsg(KHDR_ALPHAKEY_UP, vkCode, 0 /* repeatRate */, 0);
				else
					AddMsg(KHDR_KEY_UP, vkCode, 0 /* repeatRate */, 0);
				result = noErr;
				break;

			case kEventRawKeyModifiersChanged:
			{
				static UInt32 lastModifiers;
				WORD msgType = 0;

				UInt32 changed = modifiers ^ lastModifiers;
				lastModifiers = modifiers;
				if (changed & shiftKey)
				{
					keyChar = VK_SHIFT;
					msgType = (modifiers & shiftKey) ? KHDR_KEY_DOWN : KHDR_KEY_UP;
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
				if (changed & optionKey)
				{
					keyChar = VK_MENU;
					msgType = (modifiers & optionKey) ? KHDR_KEY_DOWN : KHDR_KEY_UP;
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
#if 1
				if (changed & controlKey)
				{
					keyChar = VK_CONTROL;
					msgType = (modifiers & controlKey) ? KHDR_KEY_DOWN : KHDR_KEY_UP;
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
#else
				if (changed & cmdKey)
				{
					keyChar = VK_CONTROL;
					msgType = (modifiers & cmdKey) ? KHDR_KEY_DOWN : KHDR_KEY_UP;
					AddMsg(msgType, keyChar, 0 /* repeatRate */, 0);
				}
#endif
				result = noErr;
			}
		}
	}
	else if (eventClass == kEventClassMouse)
	{
		Point where;
		(void)GetEventParameter (inEvent, kEventParamMouseLocation, typeQDPoint, NULL, sizeof(where), NULL, &where);
		GlobalToLocal (&where);
		DWORD location = (where.h & 0xffff) | ((where.v & 0xffff) << 16);

		EventMouseButton theButton;
		(void)GetEventParameter (inEvent, kEventParamMouseButton, typeMouseButton, NULL, sizeof(theButton), NULL, &theButton);
		
		switch (eventKind)
		{
			case kEventMouseDown:
				if (gInBackground || gInPrintLoop)
					sIgnoreNextMouseUp = true;
				else if (where.v >= 0)
				{
					if (theButton == 4)
						// Map button 4 to KINT's "Back"
						AddMsg(KHDR_KEY_DOWN, 0x00008001, 0, location);
					else if (theButton == 5)
						// Map button 5 to KINT's "Next"
						AddMsg(KHDR_KEY_DOWN, 0x00008002, 0, location);
					else
						AddMsg(KHDR_KEY_DOWN, theButton, 0, location);
					result = noErr;
				}
				break;
			case kEventMouseUp:
				// If we got a mouse-down while in the background, ignore the matching
				// mouse-up. This avoids sending mouse-up events to KINT when a mouse-click
				// was used to move the app to the foreground.
				if (sIgnoreNextMouseUp)
					sIgnoreNextMouseUp = false;
				else if (where.v >= 0 && !gInBackground)
				{
					if (theButton == 4)
						// Map button 4 to KINT's "Back"
						AddMsg(KHDR_KEY_UP, 0x00008001, 0, location);
					else if (theButton == 5)
						// Map button 5 to KINT's "Next"
						AddMsg(KHDR_KEY_UP, 0x00008002, 0, location);
					else
						AddMsg(KHDR_KEY_UP, theButton, 0, location);
					result = noErr;
				}
				break;
			case kEventMouseMoved:
				AddMsg(KHDR_POINTER_MOVE, where.h, where.v, 0);
				result = noErr;
				break;
			case kEventMouseDragged:
				AddMsg(KHDR_POINTER_MOVE, where.h, where.v, theButton);
				result = noErr;
				break;
			case kEventMouseWheelMoved:
				{
					EventMouseWheelAxis mouseAxis;
					(void)GetEventParameter (inEvent, kEventParamMouseWheelAxis, typeMouseWheelAxis, NULL, sizeof(mouseAxis), NULL, &mouseAxis);
					
					SInt32 mouseDelta;
					(void)GetEventParameter (inEvent, kEventParamMouseWheelDelta, typeSInt32, NULL, sizeof(mouseDelta), NULL, &mouseDelta);
					
					if (mouseAxis == kEventMouseWheelAxisY)
						AddMsg(KHDR_KEY_DOWN, 0x00008003, mouseDelta, location);
					result = noErr;
				}
				break;
		}
	}

	return result;
}
DEFINE_ONE_SHOT_HANDLER_GETTER( windowEventHandler );

//********************************************************************

void KINT_ProcessAllEvents(void)
{
	while (!KINTMsgQueue.empty())
		SendNextMsg();
}

static pascal void sIdleTimerProc(EventLoopTimerRef inTimerRef, EventLoopIdleTimerMessage inAction, void *inUserData)
{
	static int sIdleCount;
	
	if (gInPrintLoop) return;

	KINT_ProcessAllEvents();

	// If any sounds have stopped playing, dispose of them.
	while (!sSoundOrphans.empty())
	{
		movieSound_t *sound = sSoundOrphans.front();
		StopSound_QT(sound);
		sSoundOrphans.pop_front();
	}

	if (inAction == kEventLoopIdleTimerStarted)
		sIdleCount = 0;
		
	while ( KEmsg(&mainKE, KHDR_IDLE, sIdleCount++, 0, 0) ) {}
}

#pragma mark -

int main (void)
{
	EventTypeSpec windowEvents[] =
	{
		{ kEventClassWindow, kEventWindowUpdate },
		{ kEventClassWindow, kEventWindowClose },
		{ kEventClassWindow, kEventWindowCollapsed },
		{ kEventClassWindow, kEventWindowExpanded },
		{ kEventClassWindow, kEventWindowBoundsChanged },
		{ kEventClassWindow, kEventWindowHidden },
		{ kEventClassWindow, kEventWindowShown },

		{ kEventClassKeyboard, kEventRawKeyDown },
		{ kEventClassKeyboard, kEventRawKeyUp },
		{ kEventClassKeyboard, kEventRawKeyModifiersChanged },

		{ kEventClassMouse, kEventMouseDown },
		{ kEventClassMouse, kEventMouseUp },
		{ kEventClassMouse, kEventMouseMoved },
		{ kEventClassMouse, kEventMouseDragged },
		{ kEventClassMouse, kEventMouseWheelMoved },
	};
		
	EventTypeSpec appEvents[] =
	{
		{ kEventClassApplication, kEventAppActivated },
		{ kEventClassApplication, kEventAppDeactivated },
		{ kEventClassApplication, kEventAppHidden },
		{ kEventClassApplication, kEventAppShown },

		{ kEventClassCommand, kEventProcessCommand }
	};

	OSStatus err;
	IBNibRef theNib;

	// Load the menu bar from the nib
	err = CreateNibReference (CFSTR("main"), &theNib);  
	if (!err)
		SetMenuBarFromNib (theNib, CFSTR("MenuBar"));

	EnterMovies();

	//
	// Tweak the "About <game>" menu item to use the CFBundleName from our plist.
	//
	MenuRef appMenu = NULL;
	MenuItemIndex menuIndex;

	// Get the menu item and menu ref that correspond to the 'abou' command
	MenuRef rootMenu = AcquireRootMenu();
	err = GetIndMenuItemWithCommandID(rootMenu, 'abou', 1, &appMenu, &menuIndex);
	ReleaseMenu(rootMenu);
	
	// Get the existing menu item string (e.g. "About ")
	CFStringRef menuString = NULL;
	err = CopyMenuItemTextAsCFString (appMenu, menuIndex, &menuString);

	// Get the real game name string, stored as EKSGameName in the plist.
	gAppName = (CFStringRef)CFBundleGetValueForInfoDictionaryKey(CFBundleGetMainBundle(), CFSTR("EKSGameName"));
	
	// Concatenate the 2 strings and set the menu item string to the resul.
	CFStringRef newString = CFStringCreateWithFormat(NULL, NULL, CFSTR("%@%@"), menuString, gAppName);
	err = SetMenuItemTextWithCFString (appMenu, menuIndex, newString);

	if (newString)	CFRelease (newString);
	if (menuString)	CFRelease (menuString);

	// try to load the .KE file
	if( !LoadKEFile("main.ke", &mainKE) )
		exit(0);

#if _DEBUG
	DWORD majorVers = GETMEMB(mainKE.pKode+KHDR_MAJOR_INTERP_REV);
	DWORD minorVers = GETMEMB(mainKE.pKode+KHDR_MINOR_INTERP_REV);
	printf ("KINT v%d.%d\n", INTERP_MAJOR_REV, INTERP_MINOR_REV);
	printf ("%s v%d.%d\n", CFStringGetCStringPtr(gAppName, kCFStringEncodingMacRoman), majorVers, minorVers);
#endif

	SysStartup();
	
	InitScreen(0, 0);

	// call the BEGIN_STARTUP code in the KE module (tell it we're starting up and
	// about to create the main program window)
	KEmsg(&mainKE, KHDR_BEGIN_STARTUP, 0, 0, 0);

	require( CreateMacWindow(), cantCreateWindow );

    InstallApplicationEventHandler(NewEventHandlerUPP(appCommandHandler), GetEventTypeCount(appEvents), appEvents, NULL, NULL);
	InstallWindowEventHandler (gMacWindow, GetwindowEventHandlerUPP(), GetEventTypeCount(windowEvents), windowEvents, NULL, NULL);
	// Install a repeating idle timer that will fire after 100ms of inactivity.
	InstallEventLoopIdleTimer(GetCurrentEventLoop(), (kEventDurationMillisecond * 100), (kEventDurationMillisecond * 100), NewEventLoopIdleTimerUPP(sIdleTimerProc), 0, &sIdleTimerRef);

	// call the END_STARTUP code in the KE module (tell it to finish setting up, and
	// that the main program window has now been created.
	KEmsg(&mainKE, KHDR_END_STARTUP, 0, 0, 0);

	ShowWindow (gMacWindow);

	// send one IDLE message to the KASM code in case it needs to do anything once
	// right up front after everything else has been initialized.
	KEmsg(&mainKE, KHDR_IDLE, 0, 0, 0);

	ProgramReady = TRUE;

	// Run the OS X event loop
	RunApplicationEventLoop();

cantCreateWindow:
	return 0;
}

//**********************************************************************
//* Called from PMACH.CPP when the .KE wants the program to terminate
//* Do whatever's necessary to make the program terminate.
//*
void KillProgram()
{
	KEmsg(&mainKE, KHDR_SHUT_DOWN, 0, 0, 0);
	SysCleanup();
	free(mainKE.pKode);
	StopTimer();
	ExitToShell ();
}

//***********************************************************************
void KINT_SetWindowPos(DWORD r11, DWORD r12, DWORD r13, DWORD r14, DWORD r15)
{
#if _DEBUG
	printf("KINT_SetWindowPos, cmd: %08x, h: %d, w: %d, top: %d, left: %d\n", r11, r12, r13, r14, r15);
#endif

	BOOL t = InSetWindowPos;
	InSetWindowPos = TRUE;
	// We ignore the position coords and only resize. After that's done, we recenter.
	if (gMacWindow)
	{
		Rect windowBounds;
		(void)GetWindowBounds (gMacWindow, kWindowContentRgn, &windowBounds);
		
		windowBounds.right = windowBounds.left + r13;
		windowBounds.bottom = windowBounds.top + r12;

		(void)SetWindowBounds (gMacWindow, kWindowContentRgn, &windowBounds);
		RepositionWindow (gMacWindow, NULL, kWindowCenterOnMainScreen);
	}
	else
	{
#if _DEBUG
		Debugger();
#endif
	}
	InSetWindowPos = t;
}

//**************************************************************************
void KINT_GetWindowPos(DWORD *r11, DWORD *r12, DWORD *r13, DWORD *r14, DWORD *r15)
{
	if (!gMacWindow) return;
	
	Rect windowBounds;
	(void)GetWindowBounds (gMacWindow, kWindowContentRgn, &windowBounds);
	
//	*r15 = windowBounds.left; // x
//	*r14 = windowBounds.top; // y
	*r15 = 0; // x
	*r14 = 0; // y
	*r13 = windowBounds.right - windowBounds.left; // w
	*r12 = windowBounds.bottom - windowBounds.top; // h
	*r11 = 0;

#if _DEBUG
	printf("KINT_GetWindowPos, cmd: %08x, h: %d, w: %d, top: %d, left: %d\n", *r11, *r12, *r13, *r14, *r15);
#endif
}

static pascal void sTimerProc(EventLoopTimerRef inTimerRef, void *inUserData)
{
	if (!ProgramReady) return;

	AddMsg(KHDR_TIMER, 0, 0, 0);

	// Post a synthetic event to the queue - this will wake up the event loop for processing.
	OSStatus err;
	EventRef customEvent;
	err = MacCreateEvent(NULL, 'KINT', 'Time', GetCurrentEventTime(), kEventAttributeNone, &customEvent);
	if (err == noErr)
		err = PostEventToQueue(GetCurrentEventQueue(), customEvent, kEventPriorityStandard);
}

static EventLoopTimerUPP sTimerUPP = NewEventLoopTimerUPP(sTimerProc);
//******************************************************************
void StartTimer(DWORD timval)
{
	// If the period has changed, kill off the existing event loop timer.
	if (timval != sTimerPeriod)
		StopTimer();
	
	// Install a repeating millisecond timer with our new interval.
	sTimerPeriod = timval;
	if (!sTimerRef)
		InstallEventLoopTimer(GetCurrentEventLoop(), (kEventDurationMillisecond * sTimerPeriod), (kEventDurationMillisecond * sTimerPeriod), sTimerUPP, 0, &sTimerRef);
}

//******************************************************************
void StopTimer()
{
	sTimerPeriod = 0;
	if (sTimerRef)
	{
		RemoveEventLoopTimer(sTimerRef);
		sTimerRef = NULL;
	}
}

void SetPhysicalScreenSize (void)
{
	windowSizes_t windowSizes[] =
	{
		{ 640, 480 },
		{ 800, 600 },
		{ 1024, 768 },
		{ 1152, 864 },
		{ 1280, 1024 },
		{ 1600, 1200 },
		{ 0, 0 }
	};

#if FAKE_SCREEN_SIZE
	int width = FAKE_SCREEN_SIZE_W;
	int height = FAKE_SCREEN_SIZE_H;
#else
	// Get the resolution of the main display
	CGRect mainRect = CGDisplayBounds (kCGDirectMainDisplay);
	
	int height = mainRect.size.height;
	int width = mainRect.size.width;
#endif
	
	// Compensate for the menubar and the window title bar
	height -= GetMBarHeight()*2;
	
	PhysicalScreenSizeIndex = 0;
	PhysicalScreenW = 640;
	PhysicalScreenH = 480;

	// Try to find the largest windowSize that fits on this display
	int i = 0;
	do
	{
		if ((windowSizes[i].width <= width) && (windowSizes[i].height <= height))
		{
			PhysicalScreenSizeIndex = i;
			PhysicalScreenW = windowSizes[i].width;
			PhysicalScreenH = windowSizes[i].height;
		}
		i++;
	}
	while (windowSizes[i].width != 0);
	
#if _DEBUG
	printf ("PhysW: %d PhysH: %d, PhysIndex: %d\n", PhysicalScreenW, PhysicalScreenH, PhysicalScreenSizeIndex);
#endif
}

//****************************************
bool CreateMacWindow (void)
{
	bool retVal = false;
	SetPhysicalScreenSize ();

	Rect windowBounds;
	windowBounds.top = 0;
	windowBounds.left = 0;
	windowBounds.right = PhysicalScreenW;
	windowBounds.bottom = PhysicalScreenH;
	WindowAttributes attribs = kWindowCloseBoxAttribute | kWindowCollapseBoxAttribute | kWindowStandardHandlerAttribute;

	require_noerr( CreateNewWindow (kDocumentWindowClass, attribs, &windowBounds, &gMacWindow), cantCreateWindow );

	retVal = true;
	
cantGetMainDevice:
cantCreateWindow:
	return retVal;
}

//****************************************
void InitScreen(long minW, long minH)
{
#if _DEBUG
printf("InitScreen %d, %d\n", minW, minH);
#endif
	int			redraw;
	WORD		o, ow, oh;

	if( minW==-1 ) {
		ow = kBuf->logW;
		oh = kBuf->logH;
		minW = minH = 0;
	} else {
		ow = (WORD)minW;
		oh = (WORD)minH;
	}

	if( kBuf!=NULL )
	{
		DeleteBMB(kBuf);
		redraw = 1;
	}
	else redraw = 0;

	SetPhysicalScreenSize();
	
	if( PhysicalScreenW > minW ) minW = PhysicalScreenW;
	if( PhysicalScreenH > minH ) minH = PhysicalScreenH;

	kBuf = CreateBMB((WORD)minW, (WORD)minH, TRUE);
	if( NULL==kBuf ) {
		SystemMessage("Failed creating screen BITMAP");
		exit(0);
	}

	// initialize the ENTIRE screen bitmap to black
	o = kBuf->orient;
	kBuf->orient = BMO_NORMAL;
	RectBMB(kBuf, 0, 0, kBuf->physW, kBuf->physH, 0x00000000, (DWORD)-1);
	kBuf->orient = o;

	kBuf->logW = ow;
	kBuf->logH = oh;

	if (kBuf->logW == 0)
	{
		kBuf->logW = PhysicalScreenW;
		kBuf->logH = PhysicalScreenH;
	}

	if( redraw )
	{
		// call the SCREEN_CHANGE code in the KE module
		AddMsg(KHDR_SCREEN_CHANGE, 0, 0, 0);
		SendNextMsg();
	}
}

//********************************************************************
// Display the NULL-terminated message in 'pMem' using OS resources.
void SystemMessage(char *pMem)
{
	DialogRef dialog;
	DialogItemIndex itemIndex;

	CFStringRef errorString = NULL;
	require ( errorString = CFStringCreateWithCString (NULL, pMem, kCFStringEncodingMacRoman), cantCreateString );
	
	CreateStandardAlert (kAlertStopAlert, CFSTR("An error was encountered."), errorString, NULL, &dialog);
	RunStandardAlert (dialog, NULL, &itemIndex);
	
cantCreateString:
	if (errorString) CFRelease (errorString);
}

//*********************************************************************
// store the appropriate End-Of-Line sequence into buffer 'pMem' and
// return the number of bytes written to 'pMem'.
int GetSystemEOL(BYTE *pMem)
{
#if _DEBUG
	printf("GetSystemEOL\n");
#endif
	*pMem++ = 0x0a;
	return 1;
}

//***************************************
PBMB CreateBMB(WORD w, WORD h, WORD sysBM)
{
	PBMB	pX;

	pX = (PBMB)malloc(sizeof(BMB));
	if( pX!=NULL )
	{
		pX->linelen = 4*((kPixSize*w+kPixSize)/4);	// force line of pixels to take even number of DWORDS
		pX->pBits = NULL;
		pX->gWorld = NULL;
		if( !sysBM && NULL==(pX->pBits=(BYTE *)malloc(pX->linelen*h)) )
		{
			free(pX);
			pX = NULL;
		}
		else
		{
			pX->pAND = NULL;
			pX->pNext = (void *)pbFirst;
			if( pbFirst!=NULL ) pbFirst->pPrev = (void *)pX;
			pX->pPrev = NULL;
			pbFirst = pX;
			pX->physW = w;
			pX->physH = h;
			pX->clipX1 = pX->clipY1 = 0;
			pX->clipX2 = (int)w;
			pX->clipY2 = (int)h;
			pX->logW = w;
			pX->logH = h;
			pX->orient = BMO_NORMAL;
			if (sysBM)
			{
				pX->pBits=(BYTE *)malloc(pX->linelen*h);
				Rect bounds;
				unsigned long format = k32ARGBPixelFormat;
				bounds.left = bounds.top = 0;
				bounds.right = w;
				bounds.bottom = h;
				(void)NewGWorldFromPtr (&pX->gWorld, format, &bounds, NULL, NULL, 0, (Ptr)pX->pBits, pX->linelen);
			}
		}
	}
#if _DEBUG
printf("CreateBMB w: %d h: %d sys: %d, result: %08x\n", w, h, sysBM, pX);
#endif
	return pX;
}

//***************************************
void DeleteBMB(PBMB pX)
{
#if _DEBUG
printf("DeleteBMB %08x\n", pX);
#endif
	if( pX!=NULL )
	{
		if( pX->pAND!= NULL )
			free(pX->pAND);
		else if( pX->pBits != NULL )
			free(pX->pBits);
		if (pX->gWorld)
			DisposeGWorld (pX->gWorld);
		if( pX==pbFirst )
		{
			pbFirst = (PBMB)pX->pNext;
			if( pbFirst!=NULL ) pbFirst->pPrev = NULL;
		}
		else
		{
			if( pX->pNext!=NULL ) ((PBMB)pX->pNext)->pPrev = pX->pPrev;
			if( pX->pPrev!=NULL ) ((PBMB)pX->pPrev)->pNext = pX->pNext;
		}
		free(pX);
	}
}

//**********************************************************************
void InvalidateScreen(int x1, int y1, int x2, int y2)
{
#if _DEBUG
//printf("InvalidateScreen %d, %d to %d, %d (win: %08x)\n", x1, y1, x2, y2, gMacWindow);
#endif
	if (!gMacWindow) return;
	
	Rect dirtyRect;
	dirtyRect.left = x1;
	dirtyRect.right = x2;
	dirtyRect.top = y1;
	dirtyRect.bottom = y2;
	
	InvalWindowRect (gMacWindow, &dirtyRect);
}

//**********************************************************************
void FlushScreen()
{
#if _DEBUG
printf("FlushScreen\n");
#endif
	HIWindowFlush ((HIWindowRef) gMacWindow);
}

#pragma mark -

//*********************************************************
// don't allow a program to access anything above the game directory,
// so don't allow absolute paths or .. or // or such.
char	PartPath[PATHLEN];	// used ONLY by GetFileName()!!!
char *GetFileName(char *fname, char *xname) {

	char	*p;

	if( fname[0]=='\\' ) return NULL; // fail miserably if they try absolute path, only allow SUB-dirs
	strcpy(PartPath, fname);
	for(p=PartPath; *p; p++) {
		if( (*p=='.' && *(p+1)=='.')
		 || (*p=='/' && *(p+1)=='/')
		 || (*p==':')
		) return NULL;	// fail miserably if they try some tricky directory BS, only allow SUB-dirs
		if( *p=='/' ) *p = '\\';	// DOS/Windows path conversion
	}
	strcat(PartPath, xname);	// stick desired file extension on end
	return PartPath;
}

//-----------------------------------------------------------------------------
// ConvertCFStringToUnicode
//-----------------------------------------------------------------------------
// Given a CFString, convert to a UniChar, suitable for FSRef file calls.

static int ConvertCFStringToUnicode (const CFStringRef inString, UniChar *outString)
{
	*outString = 0;
	
	assert(inString);
	assert(outString);
	if (!inString || !outString) return 0;
	
	CFIndex length = CFStringGetBytes (inString, CFRangeMake(0, CFStringGetLength(inString)), kCFStringEncodingUnicode, 0, false, (UInt8*)outString, 255, NULL);
	return length;
}

string NewBuildPath(const char *pMem, BYTE inFolderType)
{
	string temp;
	string subDirString;
	string typeString;
	string partPath;
	
	partPath = pMem;

	// Replace all "\" with "/" in the incoming path
	for (size_t i; (i = partPath.find("\\")) != string::npos; partPath.replace(i, 1, "/"));

	bool noExt = inFolderType & DIR_NOEXT;
	bool isJpg = inFolderType & DIR_LOSSY; 

	switch (inFolderType & 0x0f)
	{
		case DIR_GAME:
			break;
		case DIR_UI:
			subDirString = "/UI";
			if (!noExt)
			{
				if (isJpg)
					typeString = ".jpg";
				else
					typeString = ".bmp";		
			}
			break;
		case DIR_IMAGES:
			subDirString = "/IMAGES";
			if (!noExt)
			{
				if (isJpg)
					typeString = ".jpg";
				else
					typeString = ".bmp";		
			}
			break;
		case DIR_BACKGNDS:
			subDirString = "/Backgnds";
			if (!noExt)
			{
				if (isJpg)
					typeString = ".jpg";
				else
					typeString = ".bmp";		
			}
			break;
		case DIR_MUSIC:
			subDirString = "/Music";
			if (!noExt)
				typeString = ".mid";		
			break;
		case DIR_SOUNDS:
//			subDirString = "/Sounds";
			if (!noExt)
				typeString = ".wav";		
			break;
		case DIR_GAME_IMG:
			if (!noExt)
			{
				if (isJpg)
					typeString = ".jpg";
				else
					typeString = ".bmp";		
			}
			break;
		case DIR_UI_COMMON:
			subDirString = "/CommonUI";
			if (!noExt)
			{
				if (isJpg)
					typeString = ".jpg";
				else
					typeString = ".bmp";		
			}
			break;
		default:
			break;
	}
	
	CFURLRef parentURL = NULL;
	CFURLRef eksURL = NULL;
	CFURLRef resourceURL = NULL;

	if (strcasestr (pMem, ".cfg") || strcasestr (pMem, ".sav") || strcasestr (pMem, ".csv") || strcasestr (pMem, ".dir"))
	{
		// Look in ~/Library/Application Support/Everett Kaser Software/<EKSGameName>"
		FSRef appSupportRef;
		require_noerr( FSFindFolder (kUserDomain, kApplicationSupportFolderType, true, &appSupportRef), cantFindFolder );

		// Get the CFURL and FSRef for the "Everett Kaser Software" folder, creating it if necessary.
		HFSUniStr255 name;
		name.length = ConvertCFStringToUnicode (CFSTR("Everett Kaser Software"), name.unicode);
		FSRef folderRef;
		OSStatus err = FSCreateDirectoryUnicode (&appSupportRef, name.length, name.unicode, kFSCatInfoNone, NULL, &folderRef, NULL, NULL);
		if (err == dupFNErr)
		{
			parentURL = CFURLCreateFromFSRef (NULL, &appSupportRef);
			require(parentURL, cantGetURL);
			eksURL = CFURLCreateCopyAppendingPathComponent (NULL, parentURL, CFSTR("Everett Kaser Software"), true);
		}
		else
		{
			eksURL = CFURLCreateFromFSRef (NULL, &folderRef);
		}

		require(eksURL, cantGetURL);

		FSRef eksRef;
		(void)CFURLGetFSRef(eksURL, &eksRef);

		// Now get the CFURL and FSRef for the app-specific game name portion of the path, creating if necessary.
		name.length = ConvertCFStringToUnicode (gAppName, name.unicode);
		err = FSCreateDirectoryUnicode (&eksRef, name.length, name.unicode, kFSCatInfoNone, NULL, &folderRef, NULL, NULL);
		if (err == dupFNErr)
			resourceURL = CFURLCreateCopyAppendingPathComponent (NULL, eksURL, gAppName, true);
		else
			resourceURL = CFURLCreateFromFSRef (NULL, &folderRef);

	}
	else
		resourceURL = CFBundleCopyResourcesDirectoryURL (CFBundleGetMainBundle());
	require (resourceURL, cantGetURL);

	char fullPath[1024];
	Boolean successful; successful = CFURLGetFileSystemRepresentation(resourceURL, true, (UInt8 *)fullPath, sizeof(fullPath));

	temp = fullPath;

	temp += subDirString;
	temp += "/";
	temp += partPath;
	temp += typeString;

cantCreateString:
cantCreateFSRef:
cantFindFolder:
cantGetURL:
	if (resourceURL)
		CFRelease (resourceURL);
	if (eksURL)
		CFRelease (eksURL);
	if (parentURL)
		CFRelease (parentURL);
	return temp;
}

//*********************************************************************************
// make a sub-folder of the DIR_GAME folder or a sub-folder thereof
// return non-zero if successful, else return zero if fail
int KMakeFolder(char *fname)
{
#if _DEBUG && DEBUG_FILE
	printf ("KMakeFolder: %s\n", fname);
#endif
	string newPath = NewBuildPath(fname, DIR_GAME);
	int result = mkdir(newPath.c_str(), 0777);

	// If the path already exists, don't return an error
	if (result != 0 && errno == EEXIST)
		result = 0;
	
	return result;
}

//*********************************************************************************
// Delete an empty folder (can NOT contain ANY files)
void KDelFolder(char *fname)
{
#if _DEBUG && DEBUG_FILE
	MacNotImplemented();
	printf ("KDelFolder: %s\n", fname);
#endif
}

//*********************************************************************************
// return file handle if successful, else -1
DWORD Kopen(char *fname, int oflags, BYTE dir)
{
	DWORD retVal; retVal = (DWORD)-1;
	string mode;
	string fileName;

	require(fname, badFileName);

	fileName = NewBuildPath (fname, dir);

	switch (oflags)
	{
		case O_RDBIN :		mode = "rb"; break;
		case O_WRBIN :		mode = "wb"; break;
		case O_WRBINNEW :	mode = "wb"; break;
		default:
			require (0, badoflags);
			break;
	}

	retVal = (DWORD) fopen (fileName.c_str(), mode.c_str());
	
	if (retVal == 0)
	{
		retVal = (DWORD)-1;
	}

badFileName:
badoflags:
#if _DEBUG && DEBUG_FILE
	printf("kopen oflags: %08x mode: %s, file: %s, retVal: %08x\n", oflags, mode.c_str(), fileName.c_str(), retVal);
#endif
	
	return retVal;
}

//********************************************************************************
// In KINT/KASM, 'origin' has these values (in case your language is different):
//		SEEK_SET    0
//		SEEK_CUR    1
//		SEEK_END    2
// return new file pointer position or -1 if error
DWORD Kseek(DWORD fhandle, DWORD foffset, int origin)
{
	DWORD retVal = -1;
	require (fhandle != (DWORD)-1, badHandle);
	require (fseek ((FILE*)fhandle, foffset, origin) == 0, cantSeek);
	
	retVal = ftell ((FILE*)fhandle);
	
#if _DEBUG && DEBUG_FILE
	printf("Kseek %08x offset: %d, origin: %02x retval: %08x\n", fhandle, foffset, origin, retVal);
#endif
	
badHandle:
cantSeek:
	return retVal;
}

//*********************************************************************************
void Kclose(DWORD fhandle)
{
#if _DEBUG && DEBUG_FILE
	printf("Kclose %08x\n", fhandle);
#endif
	
	require (fhandle != (DWORD)-1, badHandle);
	
	fclose ((FILE*)fhandle);
	
badHandle:
	return;
}

//********************************************************************************
//	return 1 if at EOF, 0 if not at EOF, -1 if error
DWORD Keof(DWORD fhandle)
{
	DWORD retVal = -1;
	require (fhandle != (DWORD)-1, badHandle);
	
	retVal = feof ((FILE*)fhandle);
	if (retVal)
		retVal = 1;
	
#if _DEBUG && DEBUG_FILE
	printf("Keof %08x, %d\n", fhandle, retVal);
#endif
	
badHandle:
	return retVal;
}

//******************************************************
DWORD Kread(DWORD fhandle, BYTE *pMem, DWORD numbytes)
{
	DWORD retVal = 0;
	require (fhandle != (DWORD)-1, badHandle);
	require (pMem, badPointer);
	
	retVal = fread (pMem, 1, numbytes, (FILE*)fhandle);
	
#if _DEBUG && DEBUG_FILE
	printf("Kread %08x, numbytes: %d, result: %d\n", fhandle, numbytes, retVal);
#endif
	
badHandle:
badPointer:	
	return retVal;
}

//*******************************************************
DWORD Kwrite(DWORD fhandle, BYTE *pMem, DWORD numbytes)
{
	DWORD retVal = 0;
	require (fhandle != (DWORD)-1, badHandle);
	require (pMem, badPointer);
	
	retVal = fwrite (pMem, 1, numbytes, (FILE*)fhandle);
	
#if _DEBUG && DEBUG_FILE
	printf("Kwrite %08x, numbytes: %d, result: %d\n", fhandle, numbytes, retVal);
#endif
	
badHandle:
badPointer:	
	return retVal;
}

#pragma mark -

class FileEntry
{
public:
	BYTE		attributes;
	std::string	name;

	FileEntry(const char *inName, BYTE inAttributes);
	~FileEntry(void) {}
};

typedef std::vector<FileEntry *>::iterator fileListIter_t;

class FindFileList
{
public:
	std::vector<FileEntry *> mList;
	fileListIter_t iter;

	FindFileList(void) {}
	~FindFileList(void);
};


FindFileList::~FindFileList(void)
{
	fileListIter_t iter;
	for (iter = mList.begin(); iter != mList.end(); iter++)
		delete *iter;
};


FileEntry::FileEntry(const char *inName, BYTE inAttributes)
{
	name = inName;
	attributes = inAttributes;
}


//------------------------------------------------------------------------------------
//  wildicmp
//------------------------------------------------------------------------------------
// Case-insensitive wildcard matching.
// Matches against ? and *. Returns nonzero if match found, 0 if no match.

int wildicmp(const char *string, const char *wild)
{
	const char *charP = NULL, *wildP = NULL;

	// Attempt to match the string up to the first "*"
	while ((*string) && (*wild != '*'))
	{
		// No exact or ? match, so fail
		if ((*wild != '?') && (tolower(*wild) != tolower(*string)))
			return 0;

		// Otherwise move to the next char in each string
		wild++;
		string++;
	}

	// Match the rest of the string.
	while (*string)
	{
		// See if we hit a * token
		if (*wild == '*')
		{
			// If we're at the end of the wild string, we're done and we match.
			if (!*++wild)
				return 1;

			// wild and wildP are now the char after the * token
			wildP = wild;
			// advance to the next source char after this.
			charP = string+1;
		}
		// No * token, check for exact and ? matches.
		else if ((*wild == '?') || (tolower(*wild) == tolower(*string)))
		{
			wild++;
			string++;
		}
		else
		{
			// No match. If wildP is not NULL, we've got a * search and can consume another entry
			// in "string".
			// Otherwise we'll fail the match next time through the loop due to wild being NULL.
			wild = wildP;
			string = charP++;
		}
	}

	// Consume any * tokens at the end of the string
	while (*wild == '*')
		wild++;

	// If we're past the last * token in the wild string, we'll either be NULL or we won't
	// match (i.e. the wild string was longer than the source).
	return !*wild;
}


//******************************************************************
// Search for a filename (possibly containing '*' as wild card)
DWORD Kfindfirst(char *fname, BYTE *fbuf, BYTE *attrib, BYTE dir)
{
	FSIterator iter = NULL;
	FindFileList *fileList = NULL;
	FSCatalogInfo *catInfos = NULL;
	HFSUniStr255 *names = NULL;
	const ItemCount kMaxInfo = 64;
	DWORD retVal = (DWORD)-1;
	
	string fileName = NewBuildPath (fname, dir);

#if _DEBUG && DEBUG_FILE
	printf ("find first: %s\n", fileName.c_str());
#endif

	OSStatus err;
	FSRef fsRef;

	// Make sure we're using Mac-like file names
	char macPath[1024], macName[256];
	strcpy (macPath, fileName.c_str());
	
	// If no wildcards are specified we can optimize
	bool wild = false;
	for (const char *p = macPath; *p && !wild; p++)
		if (*p == '*' || *p == '?') wild = true;

	bool removeExtension = (dir != DIR_GAME && !(dir & DIR_NOEXT));

	if (!wild)
	{
		require_noerr(FSPathMakeRef((UInt8 *)macPath, &fsRef, NULL), cantMakePathRef);
	}
	else
	{
		// Split the directory name and filename
		char *fp = strrchr(macPath, '/');
		if (fp)
		{
			strcpy(macName, fp+1);
			*(fp+1) = '\0';
		}
		else
		{
			strcpy(macName, macPath);
			strcpy(macPath, "./");
		}
		
		// Get an FSRef for the directory
		Boolean isDirectory;
		err = FSPathMakeRef((UInt8 *)macPath, &fsRef, &isDirectory);
		if (err != noErr || !isDirectory)
		{
			return -1;
		}
	}

	// Create a catalog iterator for the directory
	if (wild)
		require_noerr( FSOpenIterator(&fsRef, kFSIterateFlat, &iter), cantOpenIterator);

	// Create the linked list
	fileList = new FindFileList();
	
	// Allocate some space to receive catalog info
	catInfos = new FSCatalogInfo[kMaxInfo];
	names = new HFSUniStr255[kMaxInfo];

	// Iterate through all catalog info for items in this directory
	do
	{
		// Grab a chunk of catalog info
		ItemCount actualItems = 0;
		ItemCount index = 0;
		FSCatalogInfoBitmap infoBits = kFSCatInfoNodeFlags + kFSCatInfoAllDates + kFSCatInfoFinderInfo + kFSCatInfoDataSizes;
		if (wild)
		{
			err = FSGetCatalogInfoBulk(iter, kMaxInfo, &actualItems, NULL, infoBits, catInfos, NULL, NULL, names);
		}
		else
		{
			err = FSGetCatalogInfo(&fsRef, infoBits, &catInfos[0], &names[0], NULL, NULL);
			if (noErr == err)
			{
				err = errFSNoMoreItems;
				actualItems = 1;
			}
		}
		while ((noErr == err || errFSNoMoreItems == err) && actualItems > 0)
		{
			// Search for all files matching our search criteria
 			const HFSUniStr255 *curUniName = &names[index];
			const FSCatalogInfo *curInfo = &catInfos[index];
			
			// Convert the HFS unicode name to a MacRoman C string
			char curName[256];
			CFStringRef nameRef = CFStringCreateWithCharacters(NULL, curUniName->unicode, curUniName->length);
			if (nameRef)
			{
				if (CFStringGetCString(nameRef, curName, 256, kCFStringEncodingMacRoman))
				{
					if (wild == false || wildicmp(curName, macName))
					{
						UInt32 attribs = 0;
						if (curInfo->nodeFlags & kFSNodeIsDirectoryMask)
							attribs |= 0x10;

						if (removeExtension)
						{
							char *dotPos;
							if (dotPos = strrchr(curName, '.'))
								*dotPos = 0;
						}

						// Create an entry for this file for the linked list
						FileEntry *entry = new FileEntry(curName, attribs);

						fileList->mList.push_back(entry);
					}
				}
				CFRelease(nameRef);
			}
			--actualItems; // done with this item, process next
			++index;
		}
		// done with this chunk of catalog info, get next
	}
	while (err == noErr);
	
	// Point the list iterator to the first file entry
	fileList->iter = fileList->mList.begin();
	
	// Call Kfindnext to get the first item
	if (Kfindnext((DWORD)fileList, fbuf, attrib) == 0)
		retVal = (DWORD)fileList;

cantMakePathRef:
cantOpenIterator:
	if (iter)		FSCloseIterator(iter);
	if (catInfos)	delete[] catInfos;
	if (names)		delete[] names;

	if (fileList && retVal == (DWORD)-1)
		delete fileList;

	return retVal;
}

//********************************************************************
// Search for next filename in sequence started by Kfindfirst()
DWORD Kfindnext(DWORD ffhandle, BYTE *fbuf, BYTE *attrib)
{
	if (ffhandle)
	{
		FindFileList *fileList = (FindFileList *) ffhandle;
		
		if (fileList->iter != fileList->mList.end())
		{
			// Found a match, return the data for this entry
			*attrib = (*(fileList->iter))->attributes;
			strcpy((char *)fbuf, (*(fileList->iter))->name.c_str());

			fileList->iter++;
			
			return 0;
		}
	}

	return -1;
}

//****************************************************
// Terminate the Kfindfirst()/Kfindnext() process
void Kfindclose(DWORD ffhandle)
{
	if (ffhandle)
	{
		// Switch over to using the correct list
		FindFileList *fileList = (FindFileList *) ffhandle;

		// Dispose of our list header
		delete fileList;
	}
}

#pragma mark -

//*************************************************
void Kdelete(char *fname, BYTE dir)
{
#if _DEBUG && DEBUG_FILE
	printf ("Kdelete: %s\n", fname);
#endif
	string fileName;

	require(fname, badFileName);

	fileName = NewBuildPath (fname, dir);
	remove (fileName.c_str());

badFileName:
	return;
}

//*****************************************************
WORD KCopyFile(char *src, char *dst)
{
#if _DEBUG && DEBUG_FILE
	MacNotImplemented();
	printf ("KCopyFile: %s to %s\n", src, dst);
#endif
	return FALSE;
}

//*****************************************************
void KGetTime(KTIME *pT)
{
	CFTimeZoneRef timeZone = CFTimeZoneCopyDefault();
	CFGregorianDate curtime = CFAbsoluteTimeGetGregorianDate(CFAbsoluteTimeGetCurrent(), timeZone);
	
	if (timeZone)
		CFRelease(timeZone);

	pT->y = curtime.year;
	pT->mo= curtime.month;
	pT->d = curtime.day;
	pT->h = curtime.hour;
	pT->mi= curtime.minute;
	pT->s = curtime.second;
}

//******************************************************
void KTimeToString(BYTE *pMem)
{
	CFTimeZoneRef timeZone = CFTimeZoneCopyDefault();
	CFGregorianDate curtime = CFAbsoluteTimeGetGregorianDate(CFAbsoluteTimeGetCurrent(), timeZone);
	
	if (timeZone)
		CFRelease(timeZone);

	sprintf((char *)pMem, "%d-%d-%d %2.2d:%2.2d:%2.2d", curtime.month, curtime.day, curtime.year, curtime.hour, curtime.minute, curtime.second);
}

//****************************************************************************
//		Return a millisecond count which doesn't necessarily bear
//		any resemblance to "real" time, but rather is a counter that is
//		a reasonably accurate "millisecond counter" that may eventually
//		roll over to 0, but which should not do so more often than once a day,
//		and who's count value continues upwards (until such time as a roll-over
//		might occur).  This is accomplished on the MS-Windows platform using
//		the GetTickCount() function.  Other platforms may choose to implement
//		this in whatever way seems easiest, while meeting the above criteria.
//		This counter value is only used for timing the speed of animations
//		and brief delays in actions.  So, reasonable accuracy down to 5 or 10
//		milliseconds would be very nice, and the roll-over (if and when it
//		occurs) should only cause (at most) a slight burp in any animation
//		or screen action that is going on at the moment (but from which it
//		should recover almost immediately).
//    NOTE: It is IMPORTANT that his count be relative accurate, as it is
//		used for generating the elapsed time count for solving puzzles!
//	It could be done getting the time of day, and subtracting that from the
//  time of day when the program started, but the Windows GetTickCount() function
//	works nicely, too.

DWORD KGetTicks()
{
	static double sMillisecondsPerBeat = 0.0;
	
	// If this is the first time here, create a conversion factor
	// so we can convert the timebase registers to milliseconds.
	if (sMillisecondsPerBeat == 0.0)
	{
		mach_timebase_info_data_t timebase; 
		mach_timebase_info (&timebase);
		
		// Convert nanosecond result to milliseconds
		sMillisecondsPerBeat = (double) timebase.numer / (double) timebase.denom * 0.000001; 
	}

	// Read the timebase registers and convert to ms.
	uint64_t ticks = mach_absolute_time();
	return (DWORD)(ticks * sMillisecondsPerBeat);
}

void KSleep(DWORD tm)
{
	usleep(tm*1000);
}

//**********************************************************
void KSetCursor(WORD curtype)
{
//	curtype = 0 for normal, 1 for wait, 2 for hotlink
	ThemeCursor newCursor;
	switch (curtype)
	{
		case 0:		newCursor = kThemeArrowCursor; break;
		case 1:		newCursor = kThemeWatchCursor; break;
		case 2:		newCursor = kThemePointingHandCursor; break;
		default:	newCursor = kThemeArrowCursor; break;
	}
	SetThemeCursor (newCursor);
}

void KCaptureMouse(BOOL cap)
{
//	MacNotImplemented();
}

//**********************************************************
void KSetTitle(BYTE *pMem)
{
	require (pMem, badParam);
	require (gMacWindow, badParam);
	
	CFStringRef windowTitle; windowTitle = NULL;
	windowTitle = CFStringCreateWithCString (kCFAllocatorDefault, (const char *)pMem, kCFStringEncodingUTF8);
	require (windowTitle, cantCreateString);
	
	require_noerr( SetWindowTitleWithCFString (gMacWindow, windowTitle), cantSetWindowTitle );

badParam:
cantCreateString:
cantSetWindowTitle:
	if (windowTitle)
		CFRelease (windowTitle);
	return;
}

#pragma mark -

PMPrintSession printSession = NULL;
PMPrintSettings printSettings = NULL;
PMPageFormat pageFormat = NULL;

GrafPtr savedPrintPort = NULL;
Rect pageBounds;
void* printingContext = NULL;

short fontID = kFontIDCourier;
short fontSize = 12;
StyleParameter fontFace = 0;

//***********************************************************
// Return 0 if fail, non-0 if succeed (and set pw, ph, Xppi, Yppi)
// Get everything setup for printing which is to follow
DWORD KPrintStart(DWORD *pw, DWORD *ph, DWORD *Xppi, DWORD *Yppi)
{
#if _DEBUG && DEBUG_PRINT
	printf("KPrintStart\n");
#endif

	OSStatus status = noErr;
	DWORD retVal = 0;
	
	gInPrintLoop = true;

	require_noerr( PMCreateSession (&printSession), cantCreateSession);
	require_noerr( PMCreatePrintSettings (&printSettings), cantCreateSettings);
	require_noerr( PMSessionDefaultPrintSettings (printSession, printSettings), cantSetDefaults);

//	require_noerr( PMSetJobNameCFString (printSettings, windowTitleRef), cantSetJobName);

	require_noerr( PMCreatePageFormat (&pageFormat), cantCreatePageFormat);
	require_noerr( PMSessionDefaultPageFormat (printSession,  pageFormat), cantGetDefaultPageFormat);

	require_noerr( PMSessionSetDestination (printSession, printSettings, kPMDestinationPreview, kPMDocumentFormatPDF, NULL), cantSetDestination);

	Boolean result;
	require_noerr( PMSessionValidatePageFormat (printSession, pageFormat, &result), cantValidateFormat);

	PMResolution res;
	
	status = PMGetResolution (pageFormat, &res);
	*Xppi = (DWORD)res.hRes;
	*Yppi = (DWORD)res.vRes;

	PMRect pageRect;
	status = PMGetAdjustedPageRect (pageFormat, &pageRect);

	*pw = (pageRect.right - pageRect.left);
	*ph = (pageRect.bottom - pageRect.top) /*+ (fontSize*2)*/;
	
#if _DEBUG && DEBUG_PRINT
printf("  w: %d h: %d x ppi: %d y ppi: %d\n", *pw, *ph, *Xppi, *Yppi);
#endif

	// This will tickle the idle timer, so we have to set gInPrintLoop beforehand.
	require_noerr( PMSessionBeginDocument (printSession, printSettings, pageFormat), cantBeginSession);

	retVal = 1;
	
cantCreateSession:
cantCreateSettings:
cantSetDefaults:
cantSetJobName:
cantCreatePageFormat:
cantGetDefaultPageFormat:
cantGetContext:
cantSetDestination:
cantValidateFormat:
cantBeginSession:

	if (retVal == 0)
		KPrintStop();

	return retVal;
}

//*******************************************************
// Shutdown and clean up after printing, we're all through
void KPrintStop()
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintStop\n");
#endif
	if (pageFormat)
	{
		PMRelease (pageFormat);
		pageFormat = NULL;
	}
	if (printSettings)
	{
		PMRelease (printSettings);
		printSettings = NULL;
	}
	if (printSession)
	{
		PMSessionEndDocument (printSession);
		PMRelease (printSession);
		printSession = NULL;
	}

	gInPrintLoop = false;
}

//*****************************************************
//	return non-0 if successful, 0 if fail
BYTE KPrintBeginPage()
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintBeginPage\n");
#endif
	BYTE retVal = 0;

	require_noerr( PMSessionBeginPage(printSession, pageFormat, NULL), cantBeginPage);

	GetPort (&savedPrintPort);
	require_noerr( PMSessionGetGraphicsContext(printSession, kPMGraphicsContextQuickdraw, (void **) &printingContext), cantGetContext);
	SetPort ((CGrafPtr)printingContext);

	GetPortBounds ((CGrafPtr)printingContext, &pageBounds);

	retVal = 1;

//	QDSwapPortTextFlags((CGrafPtr)printingContext, kQDUseCGTextRendering | kQDUseCGTextMetrics);
	
cantBeginPage:
cantGetContext:

	return retVal;
}

//*****************************************************
//	return non-0 if successful, 0 if fail

BYTE KPrintEndPage()
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintEndPage\n");
#endif
	if (savedPrintPort)
	{
		SetPort (savedPrintPort);
		savedPrintPort = NULL;
	}
	if (PMSessionEndPage (printSession))
		return 0;
	else
		return 1;
}

//******************************************************
void KPrintText(int x, int y, char *str, DWORD clr)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintText x: %d y: %d %s\n", x, y, str);
#endif
	unsigned char pascalString[1024];
	
	TextFont(fontID);
	TextSize(fontSize);
	TextFace(fontFace);

	MoveTo (x, y);
	c2pstrcpy (pascalString, str);
	DrawString (pascalString);
}

//***********************************************************
void KPrintGetTextSize(char *str, DWORD *dw, DWORD *dh, DWORD *dasc)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintGetTextSize\n");
#endif
	TextFont(fontID);
	TextSize(fontSize);
	TextFace(fontFace);

	FontInfo fontInfo;
	GetFontInfo(&fontInfo);
	
	*dasc = fontInfo.ascent;
	
	*dh =  fontInfo.ascent + fontInfo.descent + fontInfo.leading;
	
	TextWidth(str, 0, strlen(str));

}

//**************************************************************************
void KPrintBMB(int xd, int yd, int wd, int hd, PBMB bmS, int xs, int ys, int ws, int hs)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintBMB\n");
#endif
//	MacNotImplemented();
}

//**************************************************************************
void KPrintRect(int xd, int yd, int wd, int hd, DWORD fillcolor, DWORD edgecolor)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintRect\n");
#endif
//	MacNotImplemented();
}

//************************************************************************************
void KPrintFont(WORD hlim, WORD flags, WORD *h, WORD *asc)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintFont\n");
#endif
	// flags:
	// bit 0: 1 = fixed-width (Win32: "Courier New" vs. "Tahoma")
	// bit 1: 1 = bold
	// bit 2: 1 = italic

	if (flags & 0x01)
		fontID = kFontIDCourier;
	else
		fontID = kFontIDHelvetica;

	if (flags & 0x02)
		fontFace |= bold;
	if (flags & 0x04)
		fontFace |= italic;

	fontSize = hlim;

	FontInfo fontInfo;
	GetFontInfo(&fontInfo);

//	*h = fontInfo.ascent + fontInfo.descent;
	*h = fontSize;
	*asc = fontInfo.ascent;
}

//*********************************************************
void KPrintLines(DWORD *pMem, WORD numpoints, DWORD clr)
{
#if _DEBUG && DEBUG_PRINT
printf("KPrintLines\n");
#endif
//	MacNotImplemented();
}

#pragma mark -
#pragma mark  Audio

static GWorldPtr		sTempGWorld;
static QTCallBackUPP	sQTCallBackUPP;
static movieSound_t		*sMusicSound;

pascal void loopCallback (QTCallBack cb, long refCon)
{
	movieSound_t *qtSound = (movieSound_t *) refCon;

	if (qtSound->mMovie)
	{
		if (qtSound->mIsMusic)
		{
			MusicPlaying = 0;
			AddMsg(KHDR_MUSIC_DONE, 0, 0, 0);
		}
		if (qtSound->mLooping)
		{
			GoToBeginningOfMovie (qtSound->mMovie);
			StartMovie (qtSound->mMovie);
			CallMeWhen (qtSound->qtCallback, sQTCallBackUPP, (long) qtSound, triggerAtStop, 0, 0);
		}
		else
		{
			qtSound->mPlaying = false;
			if (qtSound->mAutoDelete)
			{
				// Add to orphan queue for deletion in the idle timer
				sSoundOrphans.push_back(qtSound);
			}
		}
	}
}

void SetupLoopCallback(movieSound_t *inSound)
{
	if (!inSound && !inSound->mMovie) return;

	if (!sQTCallBackUPP)
		sQTCallBackUPP = NewQTCallBackUPP (loopCallback);
	
	if (!inSound->qtCallback)
		inSound->qtCallback = NewCallBack (GetMovieTimeBase (inSound->mMovie), callBackAtExtremes + callBackAtInterrupt);
	(void)CallMeWhen (inSound->qtCallback, sQTCallBackUPP, (long) inSound, triggerAtStop, 0, 0);
}

movieSound_t * PlayFile_QT (const char *inMacPath, bool inLooping, bool inDisposeWhenDone, bool inMusic)
{
	OSType dataRefType;
	FSRef fsRef;
	movieSound_t *newSound = NULL;

	require_noerr(FSPathMakeRef((UInt8 *)inMacPath, &fsRef, NULL), cantCreatePath);

	if (!sTempGWorld)
	{
		Rect r = {0,0,4,4};

		// Create a dummy gworld so we have something valid for sounds when they are created
		(void)NewGWorld (&sTempGWorld, 8, &r, NULL, NULL, keepLocal);
	}

	newSound = (movieSound_t *)calloc (sizeof(movieSound_t), 1);
	newSound->mLooping = inLooping;
	newSound->mAutoDelete = inDisposeWhenDone;
	newSound->mIsMusic = inMusic;

	// Open the movie
	require_noerr( QTNewDataReferenceFromFSRef(&fsRef, 0, &(newSound->mDataRef), &dataRefType), cantCreateDataRef);
	require_noerr( NewMovieFromDataRef(&(newSound->mMovie), newMovieDontAskUnresolvedDataRefs, NULL, newSound->mDataRef, dataRefType), cantCreateMovie);

	SetMovieGWorld (newSound->mMovie, sTempGWorld, NULL);
	
	// Go to the beginning of the movie
	SetMovieActive(newSound->mMovie, true);
	GoToBeginningOfMovie(newSound->mMovie);
	PrerollMovie(newSound->mMovie, 0, 0);

	SetupLoopCallback(newSound);
	
	// Start playback
	StartMovie(newSound->mMovie);
	newSound->mPlaying = true;

cantCreatePath:
cantCreateDataRef:
cantCreateMovie:
	if (newSound && newSound->mDataRef)
	{
		DisposeHandle(newSound->mDataRef);
		newSound->mDataRef = NULL;
	}
	return newSound;
}

void StopSound_QT (movieSound_t *inSound)
{
	if (!inSound) return;
	
	if (inSound->mMovie)
	{
		if (inSound->qtCallback)
		{
			CancelCallBack (inSound->qtCallback);
			DisposeCallBack (inSound->qtCallback);
			inSound->qtCallback = NULL;
		}
		
		if (inSound->mPlaying)
		{
			StopMovie(inSound->mMovie);
			SetMovieActive(inSound->mMovie, false);
			GoToEndOfMovie(inSound->mMovie);
			inSound->mPlaying = false;
		}
		
		DisposeMovie (inSound->mMovie);
		if (inSound->mDataRef)
			DisposeHandle (inSound->mDataRef);
			
		free (inSound);
	}	
}

//*********************************************************
void StopMusic()
{
	StopSound_QT(sMusicSound);
	sMusicSound = NULL;
	MusicPlaying = 0;
}

//****************************************************************************
void PlayMusic(char *pMem)
{
#if _DEBUG && DEBUG_SOUND
printf("PlayMusic: %s\n", pMem);
#endif
	string musicPath = NewBuildPath(pMem, DIR_MUSIC);
	StopMusic();
	sMusicSound = PlayFile_QT (musicPath.c_str(), false, false, true);
	if (sMusicSound)
		MusicPlaying = (WORD)sMusicSound->mPlaying;
}

//*******************************************************************
void KPlaySound(char *pMem)
{
#if _DEBUG && DEBUG_SOUND
printf("KPlaySound: %s\n", pMem);
#endif
	string soundPath = NewBuildPath(pMem, DIR_SOUNDS);
	(void)PlayFile_QT (soundPath.c_str(), false, true, false);
}

//*******************************************************************
void KPlaySoundMem(char *pMem, BOOL nostop)
{
	// LBO 11/18/07 - where is this used?
	MacNotImplemented();
//	sndPlaySound(pMem, SND_MEMORY | SND_ASYNC | SND_NODEFAULT | (nostop?SND_NOSTOP:0));
}


//***************
//* END OF FILE *
//***************
